@@ -34,12 +34,17 @@
 #include "gnome-vfs-cdrom.h"
 #include "gnome-vfs-filesystem-type.h"
 
+#ifdef USE_HAL
+#include "gnome-vfs-hal-mounts.h"
+#endif /* USE_HAL */
+
 static void gnome_vfs_volume_monitor_daemon_class_init (GnomeVFSVolumeMonitorDaemonClass *klass);
 static void gnome_vfs_volume_monitor_daemon_init       (GnomeVFSVolumeMonitorDaemon      *volume_monitor_daemon);
 static void gnome_vfs_volume_monitor_daemon_finalize   (GObject                          *object);
 
 static void update_fstab_drives (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon);
 static void update_mtab_volumes (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon);
+
 static void update_connected_servers (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon);
 
 typedef struct {
@@ -51,6 +56,14 @@ typedef struct {
 
 static GnomeVFSVolumeMonitorClass *parent_class = NULL;
 
+/* In the event we can't connect to the HAL daemon this boolean is TRUE
+ * and we fall back to the usual fstab/mtab monitoring. 
+ *
+ * This is also useful for maintaing the non-HAL code on a system with
+ * HAL installed.
+ */
+static gboolean dont_use_hald = TRUE;
+
 /* Locking strategy:
  *
  * Its important that the public API is threadsafe, since it will be used
@@ -140,10 +153,22 @@ connected_servers_changed (GConfClient* client,
 static void
 gnome_vfs_volume_monitor_daemon_init (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 {
-	_gnome_vfs_monitor_unix_mounts (fstab_changed,
-					volume_monitor_daemon,
-					mtab_changed,
-					volume_monitor_daemon);
+#ifdef USE_HAL
+	if (_gnome_vfs_monitor_hal_mounts_init (volume_monitor_daemon)) {
+		/* It worked, do use HAL */
+		dont_use_hald = FALSE;
+	} else {
+		/* Couldn't connect to HAL daemon, don't use HAL */
+		dont_use_hald = TRUE;
+	}
+#endif /* USE_HAL */
+
+	if (dont_use_hald) {
+		_gnome_vfs_monitor_unix_mounts (fstab_changed,
+						volume_monitor_daemon,
+						mtab_changed,
+						volume_monitor_daemon);
+	}
 
 	volume_monitor_daemon->gconf_client = gconf_client_get_default ();
 	gconf_client_add_dir (volume_monitor_daemon->gconf_client,
@@ -160,16 +185,35 @@ gnome_vfs_volume_monitor_daemon_init (GnomeVFSVolumeMonitorDaemon *volume_monito
 					 NULL);
 								       
 	
-	update_fstab_drives (volume_monitor_daemon);
-	update_mtab_volumes (volume_monitor_daemon);
+	if (dont_use_hald) {
+		update_fstab_drives (volume_monitor_daemon);
+		update_mtab_volumes (volume_monitor_daemon);
+	}
+#ifdef USE_HAL
+	else {
+		_gnome_vfs_monitor_hal_get_volume_list (volume_monitor_daemon);
+	}
+#endif /* USE_HAL */
+
 	update_connected_servers (volume_monitor_daemon);
 }
 
 void
-gnome_vfs_volume_monitor_daemon_force_probe (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
+gnome_vfs_volume_monitor_daemon_force_probe (GnomeVFSVolumeMonitor *volume_monitor)
 {
-	update_fstab_drives (volume_monitor_daemon);
-	update_mtab_volumes (volume_monitor_daemon);
+	GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon;
+	
+	volume_monitor_daemon = GNOME_VFS_VOLUME_MONITOR_DAEMON (volume_monitor);
+	
+	if (dont_use_hald) {
+		update_fstab_drives (volume_monitor_daemon);
+		update_mtab_volumes (volume_monitor_daemon);
+	}
+#ifdef USE_HAL
+	else {
+		_gnome_vfs_monitor_hal_get_volume_list (volume_monitor_daemon);
+	}
+#endif /* USE_HAL */
 	update_connected_servers (volume_monitor_daemon);
 }
 
@@ -182,15 +226,22 @@ gnome_vfs_volume_monitor_daemon_finalize (GObject *object)
 
 	volume_monitor_daemon = GNOME_VFS_VOLUME_MONITOR_DAEMON (object);
 		
-	_gnome_vfs_stop_monitoring_unix_mounts ();
+	if (dont_use_hald) {
+		_gnome_vfs_stop_monitoring_unix_mounts ();
 
-	g_list_foreach (volume_monitor_daemon->last_mtab,
-			(GFunc)_gnome_vfs_unix_mount_free, NULL);
-	g_list_free (volume_monitor_daemon->last_mtab);
+		g_list_foreach (volume_monitor_daemon->last_mtab,
+				(GFunc)_gnome_vfs_unix_mount_free, NULL);
+		g_list_free (volume_monitor_daemon->last_mtab);
 	
-	g_list_foreach (volume_monitor_daemon->last_fstab,
-			(GFunc)_gnome_vfs_unix_mount_point_free, NULL);
-	g_list_free (volume_monitor_daemon->last_fstab);	
+		g_list_foreach (volume_monitor_daemon->last_fstab,
+				(GFunc)_gnome_vfs_unix_mount_point_free, NULL);
+		g_list_free (volume_monitor_daemon->last_fstab);	
+	}
+#ifdef USE_HAL
+	else {
+		_gnome_vfs_monitor_hal_mounts_shutdown (volume_monitor_daemon);
+	}
+#endif /* USE_HAL */
 
 	gconf_client_notify_remove (volume_monitor_daemon->gconf_client,
 				    volume_monitor_daemon->connected_id);
@@ -386,6 +437,7 @@ get_icon_from_type (GnomeVFSDeviceType type)
 	return g_strdup (icon_name);
 }
 
+
 static void
 diff_sorted_lists (GList *list1, GList *list2, GCompareFunc compare,
 		   GList **added, GList **removed)
@@ -446,6 +498,7 @@ get_drive_name (GnomeVFSVolumeMonitor *volume_monitor,
 		{ "zip", N_("Zip Drive") },
 		{ "memstick", N_("Memory Stick") },
 		{ "camera", N_("Camera") },
+		{ "dvd", N_("DVD") },
 	};
 	
 	name = g_path_get_basename (mount->mount_path);
@@ -543,7 +596,7 @@ create_drive_from_mount_point (GnomeVFSVolumeMonitor *volume_monitor,
 	drive->priv->display_name = get_drive_name (volume_monitor, drive, mount);
 	
 	drive->priv->is_user_visible = TRUE;
-	drive->priv->volume = NULL;
+	drive->priv->volumes = NULL;
 
 	uri = gnome_vfs_get_uri_from_local_path (mount->mount_path);
 	mounted_volume = _gnome_vfs_volume_monitor_find_mtab_volume_by_activation_uri (volume_monitor, uri);
@@ -551,7 +604,7 @@ create_drive_from_mount_point (GnomeVFSVolumeMonitor *volume_monitor,
 
 	if (mounted_volume != NULL &&
 	    mounted_volume->priv->drive == NULL) {
-		drive->priv->volume = gnome_vfs_volume_ref (mounted_volume);
+		_gnome_vfs_drive_add_mounted_volume (drive, mounted_volume);
 		_gnome_vfs_volume_set_drive (mounted_volume, drive);
 	}
 
@@ -804,14 +857,14 @@ create_vol_from_mount (GnomeVFSVolumeMonitor *volume_monitor, GnomeVFSUnixMount
 	g_free (uri);
 
 	if (containing_drive != NULL &&
-	    containing_drive->priv->volume == NULL) {
+	    g_list_find (containing_drive->priv->volumes, vol) == NULL) {
 		/* Make sure the mounted volume for a visible drive is visible */
 		if (containing_drive->priv->is_user_visible) {
 			vol->priv->is_user_visible = 1;
 		}
 		
 		vol->priv->drive = containing_drive;
-		_gnome_vfs_drive_set_mounted_volume (containing_drive, vol);
+		_gnome_vfs_drive_add_mounted_volume (containing_drive, vol);
 	}
 
 	return vol;
