@@ -229,6 +229,14 @@ typedef struct {
 	guint server_status;
 } HttpFileHandle;
 
+typedef struct {
+	char *username;
+	char *password;
+	char *keyring;
+	char *realm;
+	enum AuthnHeaderType type;
+} HttpAuthSave;
+
 static GnomeVFSResult resolve_409 		 (GnomeVFSMethod *method,
 						  GnomeVFSURI *uri,
 						  GnomeVFSContext *context);
@@ -238,12 +246,20 @@ static void	proxy_unset_authn 		 (void);
 static gboolean invoke_callback_send_additional_headers (GnomeVFSURI *uri,
 							 GList **list);
 static gboolean invoke_callback_headers_received (HttpFileHandle *handle);
+static gboolean invoke_callback_basic_authn_fill (HttpFileHandle *handle, 
+			     			  enum AuthnHeaderType authn_which);
 static gboolean invoke_callback_basic_authn	 (HttpFileHandle *handle, 
 			     			  enum AuthnHeaderType authn_which,
-			     			  gboolean previous_attempt_failed);
+			     			  gboolean previous_attempt_failed,
+						  HttpAuthSave **auth_save);
+static gboolean invoke_callback_save_authn	 (HttpFileHandle *handle, 
+			     			  enum AuthnHeaderType authn_which,
+						  HttpAuthSave *auth_save);
 static gboolean check_authn_retry_request 	 (HttpFileHandle * http_handle,
 			   			  enum AuthnHeaderType authn_which,
-			   			  const char *prev_authn_header);
+			   			  const char *prev_authn_header,
+						  gboolean first_request,
+						  HttpAuthSave **auth_save);
 static void parse_ignore_host 			 (gpointer data,
 						  gpointer user_data);
 #ifdef ENABLE_IPV6
@@ -591,7 +607,8 @@ parse_header (HttpFileHandle *handle,
 
 static GnomeVFSResult
 get_header (GnomeVFSSocketBuffer *socket_buffer,
-	    GString *s)
+	    GString *s,
+	    GnomeVFSCancellation *cancellation)
 {
 	GnomeVFSResult result;
 	GnomeVFSFileSize bytes_read;
@@ -607,7 +624,7 @@ get_header (GnomeVFSSocketBuffer *socket_buffer,
 
 		/* ANALYZE_HTTP ("==> +get_header read"); */
 		result = gnome_vfs_socket_buffer_read (socket_buffer, &c, 1,
-				&bytes_read);
+						       &bytes_read, cancellation);
 		/* ANALYZE_HTTP ("==> -get_header read"); */
 
 		if (result != GNOME_VFS_OK) {
@@ -623,7 +640,7 @@ get_header (GnomeVFSSocketBuffer *socket_buffer,
 				char next;
 
 				result = gnome_vfs_socket_buffer_peekc (
-						socket_buffer, &next);
+						socket_buffer, &next, cancellation);
 				if (result != GNOME_VFS_OK) {
 					return result;
 				}
@@ -662,17 +679,23 @@ create_handle (GnomeVFSURI *uri,
 	GString *header_string;
 	GnomeVFSResult result;
 	guint server_status;
+	GnomeVFSCancellation *cancellation;
 
 	g_return_val_if_fail (p_handle != NULL, GNOME_VFS_ERROR_INTERNAL);
 
 	*p_handle = http_file_handle_new (socket_buffer, uri);
 
+	cancellation = NULL;
+	if (context != NULL) {
+		cancellation = gnome_vfs_context_get_cancellation(context);
+	}
+	
 	header_string = g_string_new (NULL);
 
 	ANALYZE_HTTP ("==> +create_handle");
 
 	/* This is the status report string, which is the first header.  */
-	result = get_header (socket_buffer, header_string);
+	result = get_header (socket_buffer, header_string, cancellation);
 	if (result != GNOME_VFS_OK) {
 		goto error;
 	}
@@ -689,7 +712,7 @@ create_handle (GnomeVFSURI *uri,
 
 	/* Header fetching loop.  */
 	for (;;) {
-		result = get_header (socket_buffer, header_string);
+		result = get_header (socket_buffer, header_string, cancellation);
 		if (result != GNOME_VFS_OK) {
 			break;
 		}
@@ -822,9 +845,9 @@ parse_ignore_host (gpointer data, gpointer user_data)
 {
 	gchar *hostname, *input, *netmask;
 	gboolean ip_addr = FALSE, has_error = FALSE;
-	struct in_addr host, mask;
+	struct in_addr host;
 #ifdef ENABLE_IPV6
-	struct in6_addr host6, mask6;
+	struct in6_addr host6;
 #endif
 	ProxyHostAddr *elt;
 	gint i;
@@ -849,8 +872,8 @@ parse_ignore_host (gpointer data, gpointer user_data)
 			if (*endptr != '\0' || width < 0 || width > 32) {
 				has_error = TRUE;
 			}
-			elt->mask.s_addr = htonl(~0 << width);
-			elt->addr.s_addr &= mask.s_addr;
+			elt->mask.s_addr = htonl (~0 << width);
+			elt->addr.s_addr &= elt->mask.s_addr;
 		}
 		else {
 			elt->mask.s_addr = 0xffffffff;
@@ -877,7 +900,7 @@ parse_ignore_host (gpointer data, gpointer user_data)
 				elt->mask6.s6_addr[i] = 0xff;
 			}
 			elt->mask6.s6_addr[i] = (0xff << (8 - width % 8)) & 0xff;
-			ipv6_network_addr (&elt->addr6, &mask6, &elt->addr6);
+			ipv6_network_addr (&elt->addr6, &elt->mask6, &elt->addr6);
 		}
 		else {
 			for (i = 0; i < 16; ++i) {
@@ -1190,7 +1213,8 @@ https_proxy (GnomeVFSSocket **socket_return,
 	     gchar *proxy_host,
 	     gint proxy_port,
 	     gchar *server_host,
-	     gint server_port)
+	     gint server_port,
+	     GnomeVFSCancellation *cancellation)
 {
 	/* use CONNECT to do https proxying. It goes something like this:
 	 * >CONNECT server:port HTTP/1.0
@@ -1217,7 +1241,7 @@ https_proxy (GnomeVFSSocket **socket_return,
 	gint fd;
 
 	result = gnome_vfs_inet_connection_create (&http_connection, 
-			proxy_host, proxy_port, NULL);
+			proxy_host, proxy_port, cancellation);
 
 	if (result != GNOME_VFS_OK) {
 		return result;
@@ -1230,39 +1254,39 @@ https_proxy (GnomeVFSSocket **socket_return,
 	buffer = g_strdup_printf ("CONNECT %s:%d HTTP/1.0\r\n\r\n",
 			server_host, server_port);
 	result = gnome_vfs_socket_write (http_socket, buffer, strlen(buffer),
-			&bytes);
+					 &bytes, cancellation);
 	g_free (buffer);
 
 	if (result != GNOME_VFS_OK) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		return result;
 	}
 
 	buffer = proxy_get_authn_header_for_uri (NULL); /* FIXME need uri */
 	if (buffer != NULL) {
 		result = gnome_vfs_socket_write (http_socket, buffer, 
-				strlen(buffer), &bytes);
+						 strlen(buffer), &bytes, cancellation);
 		g_free (buffer);
 	}
 
 	if (result != GNOME_VFS_OK) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		return result;
 	}
 
 	bytes = 8192;
 	buffer = g_malloc0 (bytes);
 
-	result = gnome_vfs_socket_read (http_socket, buffer, bytes-1, &bytes);
+	result = gnome_vfs_socket_read (http_socket, buffer, bytes-1, &bytes, cancellation);
 
 	if (result != GNOME_VFS_OK) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		g_free (buffer);
 		return result;
 	}
 
 	if (!parse_status (buffer, &status_code)) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		g_free (buffer);
 		return GNOME_VFS_ERROR_PROTOCOL_ERROR;
 	}
@@ -1270,7 +1294,7 @@ https_proxy (GnomeVFSSocket **socket_return,
 	result = http_status_to_vfs_result (status_code);
 
 	if (result != GNOME_VFS_OK) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		g_free (buffer);
 		return result;
 	}
@@ -1280,10 +1304,10 @@ https_proxy (GnomeVFSSocket **socket_return,
 
 	g_free (buffer);
 
-	result = gnome_vfs_ssl_create_from_fd (&ssl, fd);
+	result = gnome_vfs_ssl_create_from_fd (&ssl, fd, cancellation);
 
 	if (result != GNOME_VFS_OK) {
-		gnome_vfs_socket_close (http_socket);
+		gnome_vfs_socket_close (http_socket, cancellation);
 		return result;
 	}
 
@@ -1320,7 +1344,9 @@ connect_to_uri (
 	g_return_val_if_fail (toplevel_uri != NULL, GNOME_VFS_ERROR_INTERNAL);
 
 	if (!g_ascii_strcasecmp (gnome_vfs_uri_get_scheme (&toplevel_uri->uri), 
-				"https")) {
+				"https") ||
+	    !g_ascii_strcasecmp (gnome_vfs_uri_get_scheme (&toplevel_uri->uri), 
+				"davs")) {
 		if (!gnome_vfs_ssl_enabled ()) {
 			return GNOME_VFS_ERROR_NOT_SUPPORTED;
 		}
@@ -1349,7 +1375,7 @@ connect_to_uri (
 			*p_proxy_connect = FALSE;
 
 			result = https_proxy (&socket, proxy_host, proxy_port,
-					toplevel_uri->host_name, host_port);
+					toplevel_uri->host_name, host_port, cancellation);
 
 			g_free (proxy_host);
 			proxy_host = NULL;
@@ -1379,7 +1405,7 @@ connect_to_uri (
 
 		if (https) {
 			result = gnome_vfs_ssl_create (&ssl, 
-					toplevel_uri->host_name, host_port);
+					toplevel_uri->host_name, host_port, cancellation);
 
 			if (result != GNOME_VFS_OK) {
 				return result;
@@ -1401,7 +1427,7 @@ connect_to_uri (
 	*p_socket_buffer = gnome_vfs_socket_buffer_new (socket);
 
 	if (*p_socket_buffer == NULL) {
-		gnome_vfs_socket_close (socket);
+		gnome_vfs_socket_close (socket, cancellation);
 		return GNOME_VFS_ERROR_INTERNAL;
 	}
 
@@ -1412,7 +1438,7 @@ error:
 }
 
 static GString *
-build_request (const char * method, GnomeVFSToplevelURI * toplevel_uri, gboolean proxy_connect)
+build_request (const char * method, GnomeVFSToplevelURI * toplevel_uri, gboolean proxy_connect, gboolean force_slash_at_end)
 {
 	gchar *uri_string = NULL;
 	GString *request;
@@ -1435,6 +1461,14 @@ build_request (const char * method, GnomeVFSToplevelURI * toplevel_uri, gboolean
 						      | GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD);
 	}
 
+	if (force_slash_at_end && uri_string[strlen (uri_string) - 1] != '/') {
+		char *tmp;
+
+		tmp = uri_string;
+		uri_string = g_strconcat (tmp, "/", NULL);
+		g_free (tmp);
+	}
+	
 	/* Request line.  */
 	request = g_string_new ("");
 
@@ -1451,7 +1485,7 @@ build_request (const char * method, GnomeVFSToplevelURI * toplevel_uri, gboolean
 		g_string_append_printf (request, "Host: %s:%d\r\n",
 					toplevel_uri->host_name, toplevel_uri->host_port);
 	} else {
-		g_string_append_printf (request, "Host: %s:80\r\n",
+		g_string_append_printf (request, "Host: %s\r\n",
 					toplevel_uri->host_name);
 	}
 
@@ -1473,7 +1507,8 @@ build_request (const char * method, GnomeVFSToplevelURI * toplevel_uri, gboolean
 static GnomeVFSResult
 xmit_request (GnomeVFSSocketBuffer *socket_buffer, 
 	      GString *request, 
-	      GByteArray *data)
+	      GByteArray *data,
+	      GnomeVFSCancellation *cancellation)
 {
 	GnomeVFSResult result;
 	GnomeVFSFileSize bytes_written;
@@ -1482,7 +1517,8 @@ xmit_request (GnomeVFSSocketBuffer *socket_buffer,
 
 	/* Transmit the request headers.  */
 	result = gnome_vfs_socket_buffer_write (socket_buffer, request->str, 
-			request->len, &bytes_written);
+						request->len, &bytes_written,
+						cancellation);
 
 	if (result != GNOME_VFS_OK) {
 		goto error;
@@ -1493,26 +1529,38 @@ xmit_request (GnomeVFSSocketBuffer *socket_buffer,
 		ANALYZE_HTTP ("==> Writing data");
 		
 		result = gnome_vfs_socket_buffer_write (socket_buffer, 
-				data->data, data->len, &bytes_written);
+							data->data, data->len, &bytes_written,
+							cancellation);
 	}
 
 	if (result != GNOME_VFS_OK) {
 		goto error;
 	}
 
-	result = gnome_vfs_socket_buffer_flush (socket_buffer);	
+	result = gnome_vfs_socket_buffer_flush (socket_buffer, cancellation);	
 
 error:
 	return result;
 }
 
+static void
+http_auth_save_free (HttpAuthSave *auth_save)
+{
+	g_free (auth_save->username);
+	g_free (auth_save->password);
+	g_free (auth_save->keyring);
+	g_free (auth_save->realm);
+	g_free (auth_save);
+}
+
 static GnomeVFSResult
 make_request (HttpFileHandle **handle_return,
 	      GnomeVFSURI *uri,
 	      const gchar *method,
 	      GByteArray *data,
 	      gchar *extra_headers,
-	      GnomeVFSContext *context)
+	      GnomeVFSContext *context,
+	      gboolean force_slash_at_end)
 {
 	GnomeVFSSocketBuffer *socket_buffer;
 	GnomeVFSResult result;
@@ -1521,12 +1569,20 @@ make_request (HttpFileHandle **handle_return,
 	gboolean proxy_connect;
 	char *authn_header_request;
 	char *authn_header_proxy;
-
+	gboolean first_auth;
+	HttpAuthSave *auth_save;
+	GnomeVFSCancellation *cancellation;
+	
 	g_return_val_if_fail (handle_return != NULL, GNOME_VFS_ERROR_INTERNAL);
  	*handle_return = NULL;
 
 	ANALYZE_HTTP ("==> +make_request");
 
+	cancellation = NULL;
+	if (context != NULL) {
+		cancellation = gnome_vfs_context_get_cancellation(context);
+	}
+	
 	request 		= NULL;
 	proxy_connect 		= FALSE;
 	authn_header_request	= NULL;
@@ -1534,6 +1590,8 @@ make_request (HttpFileHandle **handle_return,
 	
 	toplevel_uri = (GnomeVFSToplevelURI *) uri;
 
+	first_auth = TRUE;
+	auth_save = NULL;
 	for (;;) {
 		GList *list;
 
@@ -1547,8 +1605,9 @@ make_request (HttpFileHandle **handle_return,
 		if (result != GNOME_VFS_OK) {
 			break;
 		}
-		
-		request = build_request (method, toplevel_uri, proxy_connect);
+
+		request = build_request (method, toplevel_uri, proxy_connect,
+					 force_slash_at_end);
 
 		authn_header_request = http_authn_get_header_for_uri (uri);
 
@@ -1592,7 +1651,7 @@ make_request (HttpFileHandle **handle_return,
 		/* Empty line ends header section.  */
 		g_string_append (request, "\r\n");
 
-		result = xmit_request (socket_buffer, request, data);
+		result = xmit_request (socket_buffer, request, data, cancellation);
 		g_string_free (request, TRUE);
 		request = NULL;
 
@@ -1608,20 +1667,38 @@ make_request (HttpFileHandle **handle_return,
 			break;
 		}
 		if ((*handle_return)->server_status == HTTP_STATUS_UNAUTHORIZED) {
-			if (! check_authn_retry_request (*handle_return, AuthnHeader_WWW, authn_header_request)) {
+			if (auth_save != NULL) {
+				http_auth_save_free (auth_save);
+				auth_save = NULL;
+			}
+			if (! check_authn_retry_request (*handle_return, AuthnHeader_WWW, authn_header_request, first_auth, &auth_save)) {
 				break;
 			}
 		} else if ((*handle_return)->server_status == HTTP_STATUS_PROXY_AUTH_REQUIRED) {
-			if (! check_authn_retry_request (*handle_return, AuthnHeader_WWW, authn_header_proxy)) {
+			if (auth_save != NULL) {
+				http_auth_save_free (auth_save);
+				auth_save = NULL;
+			}
+			if (! check_authn_retry_request (*handle_return, AuthnHeader_Proxy, authn_header_proxy, first_auth, &auth_save)) {
 				break;
 			}
 		} else {
 			break;
 		}
+		first_auth = FALSE;
 		http_file_handle_destroy (*handle_return);
 		*handle_return = NULL;
 	}
 
+	if (auth_save != NULL) {
+		invoke_callback_save_authn (*handle_return,
+					    auth_save->type,
+					    auth_save);
+		
+		http_auth_save_free (auth_save);
+		auth_save = NULL;
+	}
+	
 	g_free (authn_header_request);
 	g_free (authn_header_proxy);
 
@@ -1635,7 +1712,7 @@ make_request (HttpFileHandle **handle_return,
 	}
 	
 	if (socket_buffer != NULL) {
-		gnome_vfs_socket_buffer_destroy (socket_buffer, TRUE);
+		gnome_vfs_socket_buffer_destroy (socket_buffer, TRUE, cancellation);
 	}
 	
 	ANALYZE_HTTP ("==> -make_request");
@@ -1646,13 +1723,20 @@ static void
 http_handle_close (HttpFileHandle *handle, 
 		   GnomeVFSContext *context)
 {
+	GnomeVFSCancellation *cancellation;
+	
 	ANALYZE_HTTP ("==> +http_handle_close");
 	
+	cancellation = NULL;
+	if (context != NULL) {
+		cancellation = gnome_vfs_context_get_cancellation(context);
+	}
+	
 	if (handle != NULL) {
 		if (handle->socket_buffer) {
-			gnome_vfs_socket_buffer_flush (handle->socket_buffer);
+			gnome_vfs_socket_buffer_flush (handle->socket_buffer, cancellation);
 			gnome_vfs_socket_buffer_destroy (handle->socket_buffer,
-							 TRUE);
+							 TRUE, cancellation);
 			handle->socket_buffer = NULL;
 		}
 
@@ -1683,7 +1767,7 @@ do_open (GnomeVFSMethod *method,
 	
 	if (mode & GNOME_VFS_OPEN_READ) {
 		result = make_request (&handle, uri, "GET", NULL, NULL,
-				       context);
+				       context, FALSE);
 	} else {
 		handle = http_file_handle_new(NULL, uri); /* shrug */
 	}
@@ -1731,7 +1815,7 @@ do_create (GnomeVFSMethod *method,
 		ANALYZE_HTTP ("==> Checking to see if file exists");
 		
 		result = make_request (&handle, uri, "HEAD", NULL, NULL,
-				       context);
+				       context, FALSE);
 		http_handle_close (handle, context);
 		
 		if (result != GNOME_VFS_OK &&
@@ -1745,7 +1829,7 @@ do_create (GnomeVFSMethod *method,
 	
 	ANALYZE_HTTP ("==> Creating initial file");
 	bytes  = g_byte_array_new();
-      	result = make_request (&handle, uri, "PUT", bytes, NULL, context);
+      	result = make_request (&handle, uri, "PUT", bytes, NULL, context, FALSE);
 	http_handle_close(handle, context);
 	g_byte_array_free (bytes, TRUE);
 		
@@ -1822,7 +1906,7 @@ do_close (GnomeVFSMethod *method,
 
 		ANALYZE_HTTP ("==> doing PUT");
 		result = make_request (&new_handle, uri, "PUT", bytes, 
-				       extraheader, context);
+				       extraheader, context, FALSE);
 		g_free (extraheader);
 		http_handle_close (new_handle, context);
 	} else {
@@ -1873,10 +1957,17 @@ do_read (GnomeVFSMethod *method,
 {
 	HttpFileHandle *handle;
 	GnomeVFSResult result;
+	GnomeVFSCancellation *cancellation;
 
 	ANALYZE_HTTP ("==> +do_read");
 	DEBUG_HTTP (("+Read handle=0x%08x", (unsigned int) method_handle));
 
+	cancellation = NULL;
+	if (context != NULL) {
+		cancellation = gnome_vfs_context_get_cancellation(context);
+	}
+
+	
 	handle = (HttpFileHandle *) method_handle;
 
 	if (handle->file_info->flags & GNOME_VFS_FILE_INFO_FIELDS_SIZE) {
@@ -1887,7 +1978,7 @@ do_read (GnomeVFSMethod *method,
 	}
 
 	result = gnome_vfs_socket_buffer_read (handle->socket_buffer, buffer, 
-			num_bytes, bytes_read);
+					       num_bytes, bytes_read, cancellation);
 	
 	if (*bytes_read == 0) {
 		return GNOME_VFS_ERROR_EOF;
@@ -2165,12 +2256,55 @@ process_propfind_response(xmlNodePtr n,
 }
 
 
+static gint /* GCompareFunc */
+http_glist_find_header (gconstpointer a, gconstpointer b)
+{
+	if ( NULL != a && NULL != b) {
+		return g_ascii_strncasecmp ( (const char *)a, (const char *)b, strlen((const char *)b));
+	} else {
+		return -1;
+	}
+}
+
+static char *
+redirect_parse_response_header (GList *response_headers)
+{
+	char *header;
+	const char *uri;
+	GList *node;
+	
+	node = g_list_find_custom (response_headers, (gpointer)"Location:", http_glist_find_header);
+	for (; node != NULL 
+	     ; node = g_list_find_custom (g_list_next (node), (gpointer)"Location:", http_glist_find_header)) {
+
+		header = (char *)node->data;
+
+		/* skip through the header name */
+		uri = strchr (header, (unsigned char)':');
+
+		if (uri == NULL) {
+			continue;
+		}
+
+		uri++;
+
+		/* skip to the uri */
+		for (; *uri != '\0' 
+			&& (*uri == ' ' || *uri == '\t') 
+		     ; uri++);
+
+		return g_strdup (uri);
+	}
+	return NULL;
+}
+
 
 static GnomeVFSResult
 make_propfind_request (HttpFileHandle **handle_return,
-	GnomeVFSURI *uri,
-	gint depth,
-	GnomeVFSContext *context)
+		       GnomeVFSURI *uri,
+		       gint depth,
+		       GnomeVFSContext *context,
+		       gboolean force_slash_at_end)
 {
 	GnomeVFSResult result = GNOME_VFS_OK;
 	GnomeVFSFileSize bytes_read, num_bytes=(64*1024);
@@ -2180,6 +2314,9 @@ make_propfind_request (HttpFileHandle **handle_return,
 	xmlNodePtr cur = NULL;
 	char *extraheaders = g_strdup_printf("Depth: %d\r\n", depth);
 	gboolean found_root_node_props;
+	char *redirect_to;
+	int redirect_depth;
+	GnomeVFSURI *redirect_uri;
 
 	GByteArray *request = g_byte_array_new();
 	char *request_str = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>"
@@ -2204,8 +2341,35 @@ make_propfind_request (HttpFileHandle **handle_return,
 		http_cache_invalidate_uri_and_children (uri);
 	}
 
+	redirect_depth = 0;
+	redirect_uri = NULL;
+ redirect:
 	result = make_request (handle_return, uri, "PROPFIND", request, 
-			       extraheaders, context);
+			       extraheaders, context,
+			       (redirect_depth == 0) ? force_slash_at_end : FALSE);
+
+	/* Handle PROPFIND redirects. We often get this because we missed a / at the end
+	   of the uri for a folder */
+	if (result == GNOME_VFS_OK &&
+	    ((*handle_return)->server_status == 301 ||
+	     (*handle_return)->server_status == 302)) {
+		redirect_to = redirect_parse_response_header ((*handle_return)->response_headers);
+		redirect_depth++;
+		if (redirect_to != NULL && redirect_depth < 7) {
+			if (redirect_uri != NULL) {
+				gnome_vfs_uri_unref (redirect_uri);
+			}
+			redirect_uri = gnome_vfs_uri_new (redirect_to);
+
+			uri = redirect_uri;
+			http_handle_close (*handle_return, context);
+			*handle_return = NULL;
+			g_free (redirect_to);
+			goto redirect;
+		}
+		g_free (redirect_to);
+		result = GNOME_VFS_ERROR_TOO_MANY_LINKS;
+	}
 	
 	/* FIXME bugzilla.gnome.org 43834: It looks like some http
 	 * servers (eg, www.yahoo.com) treat PROPFIND as a GET and
@@ -2332,6 +2496,9 @@ make_propfind_request (HttpFileHandle **handle_return,
 	}
 
 cleanup:
+	if (redirect_uri != NULL) {
+		gnome_vfs_uri_unref (redirect_uri);
+	}
 	g_free(buffer);
 	g_free(extraheaders);
 	g_byte_array_free (request, TRUE);
@@ -2391,20 +2558,18 @@ do_open_directory(GnomeVFSMethod *method,
 		handle->files = child_file_info_cached_list;
 		result = GNOME_VFS_OK;
 	} else {
-		result = make_propfind_request(&handle, uri, 1, context);
-		/* mfleming -- is this necessary?  Most DAV server's I've seen don't have the horrible
-		 * lack-of-trailing-/-is-a-301 problem for PROPFIND's
-		 */
-		if (result == GNOME_VFS_ERROR_NOT_FOUND) { /* 404 not found */
-			if (uri->text != NULL && *uri->text != '\0'
-			   && uri->text[strlen (uri->text) - 1] != '/') {
-				GnomeVFSURI *tmpuri = gnome_vfs_uri_append_path (uri, "/");
-				result = do_open_directory (method, (GnomeVFSMethodHandle **)&handle, tmpuri, options, context);
-				gnome_vfs_uri_unref (tmpuri);
+		/* Make sure we have a slash on the end of the uri for the
+		   folder PROPFIND request, see bugzilla #92908.
+		   This is not needed since we handle redirects, but it means
+		   we send less PROPFIND requests. */
+		result = make_propfind_request(&handle, uri, 1, context, TRUE);
 
-			}
+		/* Work around the case where there was a non-directory existing
+		   with that name and we got a 301 redirect, due to the slash
+		   added above */
+		if (result == GNOME_VFS_ERROR_NOT_SUPPORTED) {
+			result = GNOME_VFS_ERROR_NOT_A_DIRECTORY;
 		}
-
 		if (result == GNOME_VFS_OK
 		    && handle->file_info->type != GNOME_VFS_FILE_TYPE_DIRECTORY) {
 			result = GNOME_VFS_ERROR_NOT_A_DIRECTORY;
@@ -2508,7 +2673,7 @@ do_get_file_info (GnomeVFSMethod *method,
 		 * Start off by making a PROPFIND request.  Fall back to a HEAD if it fails
 		 */
 		
-		result = make_propfind_request (&handle, uri, 0, context);
+		result = make_propfind_request (&handle, uri, 0, context, FALSE);
 		
 		/* Note that theoretically we could not bother with this request if we get a 404 back,
 		 * but since some servers seem to return wierd things on PROPFIND (mostly 200 OK's...)
@@ -2531,7 +2696,7 @@ do_get_file_info (GnomeVFSMethod *method,
 
 				ANALYZE_HTTP ("==> do_get_file_info: do GET ");
 
-				result = make_request (&handle, uri, "GET", NULL, NULL, context);
+				result = make_request (&handle, uri, "GET", NULL, NULL, context, FALSE);
 				if (result == GNOME_VFS_OK) {
 					gnome_vfs_file_info_copy (file_info, handle->file_info);
 					http_cache_add_uri (uri, handle->file_info, FALSE);
@@ -2628,7 +2793,7 @@ do_make_directory (GnomeVFSMethod *method,
 	 * So we do a PROPFIND first to find out
 	 */
 	/* FIXME check cache here */
-	result = make_propfind_request(&handle, uri, 0, context);
+	result = make_propfind_request(&handle, uri, 0, context, FALSE);
 
 	if (result == GNOME_VFS_OK) {
 		result = GNOME_VFS_ERROR_FILE_EXISTS;
@@ -2638,7 +2803,7 @@ do_make_directory (GnomeVFSMethod *method,
 		
 		if (result == GNOME_VFS_ERROR_NOT_FOUND) {
 			http_cache_invalidate_uri_parent (uri);
-			result = make_request (&handle, uri, "MKCOL", NULL, NULL, context);
+			result = make_request (&handle, uri, "MKCOL", NULL, NULL, context, FALSE);
 		}
 	}
 	http_handle_close (handle, context);
@@ -2671,7 +2836,7 @@ do_remove_directory(GnomeVFSMethod *method,
 	 * directory is not empty
 	 */
 	result = make_request (&handle, uri, "DELETE", NULL, NULL,
-			       context);
+			       context, FALSE);
 	http_handle_close (handle, context);
 	
 	DEBUG_HTTP (("-Remove_Directory (%d)", result));
@@ -2722,7 +2887,7 @@ do_move (GnomeVFSMethod *method,
 	destpath = gnome_vfs_uri_to_string (new_uri, GNOME_VFS_URI_HIDE_USER_NAME|GNOME_VFS_URI_HIDE_PASSWORD);
 	destheader = g_strdup_printf ("Destination: %s\r\nOverwrite: %c\r\n", destpath, force_replace ? 'T' : 'F' );
 
-	result = make_request (&handle, old_uri, "MOVE", NULL, destheader, context);
+	result = make_request (&handle, old_uri, "MOVE", NULL, destheader, context, FALSE);
 	http_handle_close (handle, context);
 	g_free (destheader);
 	handle = NULL;
@@ -3005,37 +3170,40 @@ invoke_callback_send_additional_headers (GnomeVFSURI  *uri,
 }
 
 static gboolean
-invoke_callback_basic_authn (HttpFileHandle *handle, 
-			     enum AuthnHeaderType authn_which,
-			     gboolean previous_attempt_failed)
+invoke_callback_basic_authn_fill (HttpFileHandle *handle, 
+				  enum AuthnHeaderType authn_which)
 {
-	GnomeVFSModuleCallbackAuthenticationIn in_args;
-	GnomeVFSModuleCallbackAuthenticationOut out_args;
+	GnomeVFSModuleCallbackFillAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFillAuthenticationOut out_args;
 	gboolean ret;
+	char *username;
+	
 
 	ret = FALSE;
 	
 	memset (&in_args, 0, sizeof (in_args));
 	memset (&out_args, 0, sizeof (out_args));
 
-	in_args.previous_attempt_failed = previous_attempt_failed;
-		
 	in_args.uri = gnome_vfs_uri_to_string (handle->uri, GNOME_VFS_URI_HIDE_NONE);
+	in_args.server = (char *)gnome_vfs_uri_get_host_name (handle->uri);
+	in_args.port = gnome_vfs_uri_get_host_port (handle->uri);
+	username = (char *)gnome_vfs_uri_get_user_name (handle->uri);
+	if (username != NULL && username[0] != 0) {
+		in_args.username = username;
+	}
+	in_args.protocol = "http";
+	in_args.authtype = authn_which == AuthnHeader_WWW ? "basic" : "proxy";
 
-	ret = http_authn_parse_response_header_basic (authn_which, handle->response_headers, &in_args.realm);
+	ret = http_authn_parse_response_header_basic (authn_which, handle->response_headers, &in_args.object);
 		
 	if (!ret) {
 		goto error;
 	}
 
-	DEBUG_HTTP (("Invoking %s authentication callback for uri %s",
+	DEBUG_HTTP (("Invoking %s fill authentication callback for uri %s",
 		authn_which == AuthnHeader_WWW ? "basic" : "proxy", in_args.uri));
 
-	in_args.auth_type = AuthTypeBasic;
-
-	ret = gnome_vfs_module_callback_invoke (authn_which == AuthnHeader_WWW 
-						? GNOME_VFS_MODULE_CALLBACK_AUTHENTICATION
-						: GNOME_VFS_MODULE_CALLBACK_HTTP_PROXY_AUTHENTICATION, 
+	ret = gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
 						&in_args, sizeof (in_args), 
 						&out_args, sizeof (out_args)); 
 
@@ -3043,8 +3211,8 @@ invoke_callback_basic_authn (HttpFileHandle *handle,
 		DEBUG_HTTP (("No callback registered"));
 		goto error;
 	}
-
-	ret = (out_args.username != NULL);
+	
+	ret = out_args.valid;
 
 	if (!ret) {
 		DEBUG_HTTP (("No username provided by callback"));
@@ -3060,13 +3228,156 @@ invoke_callback_basic_authn (HttpFileHandle *handle,
 	}
 error:
 	g_free (in_args.uri);
-	g_free (in_args.realm);
+	g_free (in_args.object);
 	g_free (out_args.username);
+	g_free (out_args.domain);
 	g_free (out_args.password);
 
 	return ret;
 }
 
+
+static gboolean
+invoke_callback_basic_authn (HttpFileHandle *handle, 
+			     enum AuthnHeaderType authn_which,
+			     gboolean previous_attempt_failed,
+			     HttpAuthSave **auth_save_out)
+{
+	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
+	gboolean ret;
+	char *username;
+	HttpAuthSave *auth_save;
+
+	ret = FALSE;
+
+	*auth_save_out = NULL;
+	memset (&in_args, 0, sizeof (in_args));
+	memset (&out_args, 0, sizeof (out_args));
+
+	in_args.uri = gnome_vfs_uri_to_string (handle->uri, GNOME_VFS_URI_HIDE_NONE);
+	in_args.server = (char *)gnome_vfs_uri_get_host_name (handle->uri);
+	in_args.port = gnome_vfs_uri_get_host_port (handle->uri);
+	username = (char *)gnome_vfs_uri_get_user_name (handle->uri);
+	if (username != NULL && username[0] != 0) {
+		in_args.username = username;
+	}
+	in_args.protocol = "http";
+	in_args.authtype = authn_which == AuthnHeader_WWW ? "basic" : "proxy";
+
+	in_args.flags = GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_PASSWORD | GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_SAVING_SUPPORTED;
+	if (in_args.username == NULL) {
+		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME;
+	}	
+	if (previous_attempt_failed) {
+		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_PREVIOUS_ATTEMPT_FAILED;
+	}
+	
+	in_args.default_user = in_args.username;
+	
+	ret = http_authn_parse_response_header_basic (authn_which, handle->response_headers, &in_args.object);
+
+	if (!ret) {
+		goto error;
+	}
+
+	DEBUG_HTTP (("Invoking %s authentication callback for uri %s",
+		authn_which == AuthnHeader_WWW ? "basic" : "proxy", in_args.uri));
+
+	ret = gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION,
+						&in_args, sizeof (in_args), 
+						&out_args, sizeof (out_args)); 
+
+	if (!ret) {
+		DEBUG_HTTP (("No callback registered"));
+		goto error;
+	}
+
+	ret = !out_args.abort_auth;
+
+	if (!ret) {
+		DEBUG_HTTP (("No username provided by callback"));
+		goto error;
+	}
+
+	DEBUG_HTTP (("Back from authentication callback, adding credentials"));
+
+	username = (char *)gnome_vfs_uri_get_user_name (handle->uri);
+	if (username == NULL || username[0] == 0) {
+		username = out_args.username;
+	}
+
+	if (out_args.save_password) {
+		auth_save = g_new (HttpAuthSave, 1);
+		*auth_save_out = auth_save;
+		auth_save->username = g_strdup (username);
+		auth_save->password = g_strdup (out_args.password);
+		auth_save->keyring = g_strdup (out_args.keyring);
+		auth_save->realm = g_strdup (in_args.object);
+		auth_save->type = authn_which;
+	}
+	
+	if (authn_which == AuthnHeader_WWW) {
+		http_authn_session_add_credentials (handle->uri, username, out_args.password);
+	} else /* if (authn_which == AuthnHeader_Proxy) */ {
+		proxy_set_authn (username, out_args.password);
+	}
+error:
+	g_free (in_args.uri);
+	g_free (in_args.object);
+	g_free (out_args.username);
+	g_free (out_args.domain);
+	g_free (out_args.password);
+	g_free (out_args.keyring);
+
+	return ret;
+}
+
+static gboolean
+invoke_callback_save_authn (HttpFileHandle *handle, 
+			    enum AuthnHeaderType authn_which,
+			    HttpAuthSave *auth_save)
+{
+	GnomeVFSModuleCallbackSaveAuthenticationIn in_args;
+	GnomeVFSModuleCallbackSaveAuthenticationOut out_args;
+	gboolean ret;
+
+	ret = FALSE;
+
+	memset (&in_args, 0, sizeof (in_args));
+	memset (&out_args, 0, sizeof (out_args));
+
+	in_args.keyring = auth_save->keyring;
+
+	in_args.uri = gnome_vfs_uri_to_string (handle->uri, GNOME_VFS_URI_HIDE_NONE);
+	in_args.server = (char *)gnome_vfs_uri_get_host_name (handle->uri);
+	in_args.port = gnome_vfs_uri_get_host_port (handle->uri);
+	in_args.username = auth_save->username;
+	in_args.password = auth_save->password;
+	in_args.protocol = "http";
+	in_args.authtype = authn_which == AuthnHeader_WWW ? "basic" : "proxy";
+	in_args.object = auth_save->realm;
+
+	DEBUG_HTTP (("Invoking %s authentication save callback for uri %s",
+		authn_which == AuthnHeader_WWW ? "basic" : "proxy", in_args.uri));
+
+	ret = gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
+						&in_args, sizeof (in_args), 
+						&out_args, sizeof (out_args)); 
+
+	if (!ret) {
+		DEBUG_HTTP (("No callback registered"));
+		goto error;
+	}
+
+error:
+	g_free (in_args.uri);
+
+	return ret;
+}
+
+
+
 static int
 strcmp_allow_nulls (const char *s1, const char *s2)
 {
@@ -3084,11 +3395,14 @@ strcmp_allow_nulls (const char *s1, const char *s2)
 gboolean
 check_authn_retry_request (HttpFileHandle * http_handle,
 			   enum AuthnHeaderType authn_which,
-			   const char *prev_authn_header)
+			   const char *prev_authn_header,
+			   gboolean first_request,
+			   HttpAuthSave **auth_save)
 {
 	gboolean ret;
 	char *current_authn_header;
 
+	*auth_save = NULL;
 	current_authn_header = NULL;
 	
 	g_mutex_lock (gl_mutex);
@@ -3103,7 +3417,13 @@ check_authn_retry_request (HttpFileHandle * http_handle,
 
 	ret = FALSE;
 	if (0 == strcmp_allow_nulls (current_authn_header, prev_authn_header)) {
-		ret = invoke_callback_basic_authn (http_handle, authn_which, prev_authn_header == NULL);
+		if (first_request) {
+			ret = invoke_callback_basic_authn_fill (http_handle, authn_which);
+		}
+		if (!ret) {
+			/* TODO: shouldn't say previous_attempt_failed if the fill callback failed */
+			ret = invoke_callback_basic_authn (http_handle, authn_which, prev_authn_header == NULL, auth_save);
+		}
 	} else {
 		ret = TRUE;
 	}
